unit GsvHeap;

{   TGsvHeap  ,    
     ,      
  .       
       (  
  Delphi)   ( 64 ).   
       , ,   
   ,  ,     
   .NET.     -  
        .  
         
            
    .

     .

         
   .     ,   
  ,    . ,    ,
    .     
  :
    )        (64 ),
    )         (MinSize),
           .   0
         .

   MinSize      
         .
  MinSize        -
        .   , 
   MinSize    .

   Align       1, 2, 4 
  8        .

   Boost      
  .         
     .    
          . 
         ,
     -  .

   TotalSize      , 
    ,  ,   .

     TGsvHeap -  Alloc
    Alloc(aSize: Cardinal): Pointer
          ,
      Align.    
   .

   AllocCopy      ,   
      .

        -
    .

       -   .
}

interface

uses
  Windows, Classes;

type
  TGsvHeap = class
  private
    FList:        TList;    //  
    FMinSize:     Cardinal; //   
    FAlign:       Cardinal; // : 1, 2, 4  8 
    FBoost:       Boolean;  //   
    FFreePointer: PChar;    //      
    FFreeSize:    Cardinal; //   
    FTotalSize:   Cardinal; //   

    function  AdjustBlockSize(aSize: Cardinal): Cardinal;
    procedure AdjustAlign(aAlign: Cardinal);
    procedure AllocBlock(aSize: Cardinal);
    procedure FreeBlocks;

  protected
    procedure SetMinSize(Value: Cardinal); virtual;
    procedure SetAlign(Value: Cardinal); virtual;

  public
    constructor Create(aMinSize: Cardinal = 0; aAlign: Cardinal = 4;
                aBoost: Boolean = False);
    destructor  Destroy; override;

    procedure Clear; virtual;
    function  Alloc(aSize: Cardinal): Pointer; virtual;
    function  AllocCopy(aSize: Cardinal; const aData: Pointer): Pointer; virtual;

    function  Copy(const aData: PChar): PChar; overload; virtual;
    function  Copy(const aData: string): PChar; overload; virtual;
    function  Copy(const aData: PWideChar): PWideChar; overload; virtual;
    function  Copy(const aData: WideString): PWideChar; overload; virtual;

    property MinSize: Cardinal read FMinSize write SetMinSize;
    property Align: Cardinal read FAlign write SetAlign;
    property Boost: Boolean read FBoost write FBoost;
    property TotalSize: Cardinal read FTotalSize;
  end;

  //   
  TGsvHeapMT = class(TGsvHeap)
  private
    FLock: TRTLCriticalSection;

  protected
    procedure SetMinSize(Value: Cardinal); override;
    procedure SetAlign(Value: Cardinal); override;

  public
    constructor Create(aMinSize: Cardinal = 0; aAlign: Cardinal = 4;
                aBoost: Boolean = False);
    destructor  Destroy; override;

    procedure Clear; override;
    function  Alloc(aSize: Cardinal): Pointer; override;
    function  AllocCopy(aSize: Cardinal; const aData: Pointer): Pointer;
              override;
  end;

implementation

uses
  SysUtils;

const
  GRANULARITY = 1024 * 64; //     Windows

{ TGsvHeap }

function TGsvHeap.AdjustBlockSize(aSize: Cardinal): Cardinal;
begin
  if aSize <= GRANULARITY then
    Result := GRANULARITY
  else
    Result := (aSize + (GRANULARITY - 1)) and not (GRANULARITY - 1);
end;

procedure TGsvHeap.AdjustAlign(aAlign: Cardinal);
var
  d: Cardinal;
begin
  case aAlign of
    2, 4, 8: ;
    else     aAlign := 1;
  end;
  if FAlign <> aAlign then begin
    FAlign := aAlign;
    if Assigned(FFreePointer) and (FAlign <> 1) then begin
      //       
      d := (FAlign - (Cardinal(FFreePointer) and not (FAlign - 1))) and
            not (FAlign - 1);
      Inc(FFreePointer, d);
      Dec(FFreeSize, d);
    end;
  end;
end;

procedure TGsvHeap.AllocBlock(aSize: Cardinal);
var
  sz: Cardinal;
  p:  Pointer;
begin
  sz := AdjustBlockSize(aSize);
  if FBoost then begin
    //   
    if sz < FTotalSize then
      sz := FTotalSize;
  end
  else begin
    //   
    if sz < FMinSize then
      sz := FMinSize;
  end;
  p := VirtualAlloc(nil, sz, MEM_RESERVE or MEM_COMMIT, PAGE_READWRITE);
  if not Assigned(p) then
    OutOfMemoryError;
  FList.Add(p);
  Inc(FTotalSize, sz);
  FFreePointer := p;
  FFreeSize    := sz;
end;

procedure TGsvHeap.FreeBlocks;
var
  i: Integer;
begin
  for i := 0 to Pred(FList.Count) do begin
    VirtualFree(FList.Items[i], 0, MEM_RELEASE);
    FList.Items[i] := nil;
  end;
  FList.Clear;
  FFreePointer := nil;
  FFreeSize    := 0;
  FTotalSize   := 0;
end;

procedure TGsvHeap.SetMinSize(Value: Cardinal);
begin
  FMinSize := AdjustBlockSize(Value);
end;

procedure TGsvHeap.SetAlign(Value: Cardinal);
begin
  AdjustAlign(Value);
end;

constructor TGsvHeap.Create(aMinSize, aAlign: Cardinal; aBoost: Boolean);
begin
  FMinSize := AdjustBlockSize(aMinSize);
  FBoost   := aBoost;
  FList    := TList.Create;
  AdjustAlign(aAlign);
end;

destructor TGsvHeap.Destroy;
begin
  FreeBlocks;
  FList.Free;
  inherited Destroy;
end;

procedure TGsvHeap.Clear;
begin
  FreeBlocks;
end;

function TGsvHeap.Alloc(aSize: Cardinal): Pointer;
begin
  Assert(aSize <> 0);
  //     
  aSize := (aSize + (FAlign - 1)) and not (FAlign - 1);
  if aSize > FFreeSize then
    AllocBlock(aSize);
  Result := FFreePointer;
  Inc(FFreePointer, aSize);
  Dec(FFreeSize, aSize);
end;

function TGsvHeap.AllocCopy(aSize: Cardinal; const aData: Pointer): Pointer;
begin
  Result := Alloc(aSize);
  CopyMemory(Result, aData, aSize);
end;

function TGsvHeap.Copy(const aData: PChar): PChar;
begin
  Result := PChar(AllocCopy(StrLen(aData) + 1, aData));
end;

function TGsvHeap.Copy(const aData: String): PChar;
begin
  Result := PChar(AllocCopy(Length(aData) + 1, PChar(aData)));
end;

function TGsvHeap.Copy(const aData: PWideChar): PWideChar;
begin
  Result := PWideChar(AllocCopy((lstrlenW(aData) + 1) * 2, aData));
end;

function TGsvHeap.Copy(const aData: WideString): PWideChar;
begin
  Result := PWideChar(AllocCopy((Length(aData) + 1) * 2, PWideChar(aData)));
end;

{ TGsvHeapMT }

procedure TGsvHeapMT.SetMinSize(Value: Cardinal);
begin
  EnterCriticalSection(FLock);
  inherited SetMinSize(Value);
  LeaveCriticalSection(FLock);
end;

procedure TGsvHeapMT.SetAlign(Value: Cardinal);
begin
  EnterCriticalSection(FLock);
  inherited SetAlign(Value);
  LeaveCriticalSection(FLock);
end;

constructor TGsvHeapMT.Create(aMinSize, aAlign: Cardinal; aBoost: Boolean);
begin
  inherited Create(aMinSize, aAlign, aBoost);
  InitializeCriticalSection(FLock);
end;

destructor TGsvHeapMT.Destroy;
begin
  DeleteCriticalSection(FLock);
  inherited Destroy;
end;

procedure TGsvHeapMT.Clear;
begin
  EnterCriticalSection(FLock);
  try
    inherited Clear;
  finally
    LeaveCriticalSection(FLock);
  end;
end;

function TGsvHeapMT.Alloc(aSize: Cardinal): Pointer;
begin
  EnterCriticalSection(FLock);
  try
    Result := inherited Alloc(aSize);
  finally
    LeaveCriticalSection(FLock);
  end;
end;

function TGsvHeapMT.AllocCopy(aSize: Cardinal; const aData: Pointer):
  Pointer;
begin
  EnterCriticalSection(FLock);
  try
    Result := inherited AllocCopy(aSize, aData);
  finally
    LeaveCriticalSection(FLock);
  end;
end;

end.
